(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_io.ml 309 2010-10-30 20:20:37Z gerd $ *)

open Plasma_rpcapi_aux
open Uq_engines.Operators

class type record_config =
object
  method bigblock_size : int
  method mr_buffer_size : int
  method mr_buffer_size_tight : int
end


class type record_reader =
object
  method pos_in : int
  method input_record : unit -> string
  method peek_record : unit -> string
  method close_in : unit -> unit
  method abort : unit -> unit
  method to_fd_e : Unix.file_descr -> unit Uq_engines.engine
  method plasma_cluster : Plasma_client.plasma_cluster
end


class type record_writer =
object
  method pos_out : int
  method bpos_out : int64
  method output_record : string -> unit
  method flush : unit -> unit
  method from_fd_e : Unix.file_descr -> unit Uq_engines.engine
  method close_out : unit -> unit
  method abort : unit -> unit
  method plasma_cluster : Plasma_client.plasma_cluster
end


let zero_mem =
  Bigarray.Array1.create Bigarray.char Bigarray.c_layout 0

(** rounds the value m up to the next multiple of n *)
let round_up m n =
  ((m-1) / n + 1) * n

let bigblock_size c suggested =
  let bsize = Plasma_client.blocksize c in
  round_up suggested bsize


let read_file c rc name index len : record_reader =
  (* Turns out we cannot use c in synchronous mode (to_fd_e). Because
     of this we create a copy of c and use this for getting file info.
     Not especially elegant, but a solution.
   *)
  let () =
    if index < 0L || len < 0L then invalid_arg "Mapred_io.read_file" in
  let esys = Plasma_client.event_system c in

  (* This block of variables is initialized in buf_init_e: *)
  let bsize = ref 0 in
  let bigblocksize = ref 0 in
  let inode = ref 0L in
  let ii = ref (Plasma_client.regular_ii c 0) in
  let end_index = ref 0L in     (* the end of the requested region as index *)
  let hard_end_index = ref 0L in(* the end of the overflow region as index *)
  let buf_num = ref 0 in        (* length of bufs in number of blocks *)
  let buf_dim = ref 0 in        (* length of bufs in bytes *)
  let bufs = ref zero_mem in    (* The buffer as bigarray *)
  let buf_fd = ref Unix.stdin in (* The buffer as descriptor *)
  let shm_undo = ref (fun () -> ()) in    (* deregisters the buffer in shm *)

  (* These are initialized by the first fill_buffers_e call: *)
  let buf_end = ref 0 in        (* end position in bufs in bytes *)
  let buf_count = ref 0 in      (* how many blocks we have *)
  let buf_eof = ref None in     (* the (logical) eof position if reached *)
  let buf_index = ref index in  (* the block the first buffer holds *)
  let buf_cursor = ref 0 in     (* as byte position in bufs *)
  let buf_eof_reached = ref false in

  (** So what are we going to do? The user requests:
   - to read the blocks [index] to [index+len-1] (the "requested region")
   - and if the last line does not end in this block region, the remaining
     part of the last line is taken from the following blocks (the
     "overflow region")

   Also, it is assumed that the unit of processing are bigblocks, i.e.
   a whole multiple of blocks. So [index] must point to the beginning of
   a bigblock, and [len] must be a multiple of bigblocks (unless we are
   at the end of the file, and the last bigblock exists only partially).
   Note that there is the corner case that the last byte of the requested region
   is an LF char. Then the last line to read is the following line, i.e.
   this line is solely taken from the overflow region.

   We allocate buffers, and each of the buffers is one block. The number of
   buffers [buf_limit] is limited by [mem_limit] - but at least there are
   as many buffers so a bigblock fits in. Note that we do not require that
   [buf_limit] is a multiple of bigblocks.

   Reading is now done by: filling the buffers, taking line by line from
   the filled buffers, and if the end of the buffers is reached, the buffers
   are refilled. This is done until the complete requested region is read, and
   the part of the overflow region that is needed for getting
   the last line.

   At refill time the current line is usually split in a first part that is
   taken from the buffers before refill, and a trailing part that is taken
   from the buffers after refill. The maximum line length is [bigblock_size],
   and we have at least as many buffers to hold a whole line. This means
   it is ensured that at most one refill is needed to get the remaining part
   of a line.

   The [end_index] is the file index (i.e. file position in units of
   blocks) that marks the end of the requested region. [end_index] is
   also the beginnning of the overflow region. (Note that this is only
   in units of blocks - when computing the byte position of the end
   of the requested region, one also has to take into account that the
   last block might not be fully used.)

   The [hard_end_index] is the end of the overflow region, again in 
   blocks.

   Every refill sets:
    - [buf_index]: the index of the first buffered block
    - [buf_count]: the number of buffered blocks
    - [buf_end]: the number of buffered bytes (important if the last
      buffered block is the last block of the file which might be incomplete)
    - [buf_eof]: this is [None] if the end of the requested region is not
      yet reached. It is [Some p] when the end is reached, and the byte [p]
      of the current buffers (counting from 0 to [buf_end]) is the first
      byte of the overflow region. [Some 0] can also mean that the first
      byte of the overflow region was already reached before the last
      refill. [Some end_buf] means that either the end of the file is
      reached or when [end_buf=buf_dim] that the overflow region starts
      at the first byte after the current buffers.

   Note that the overflow region can be empty.

   While extracting lines from the buffers, the value [buf_cursor] is
   the current byte position in the buffer (ranging from 0 to [buf_end]),
   and [buf_eof_reached] is set when the last line is returned ot the
   user.

 *)

  let lazy_init = ref false in        (* whether lazy_init_e is done *)
  let lazy_init_started = ref None in
  let pos_in = ref 0 in               (* number of lines read so far *)
  let cur_rec = ref None in           (* currently peeked lines *)
  let closed = ref false in           (* whether closed *)
  let to_fd_engine = ref None in      (* whether a to_fd_e is active *)

  let find_lf pos end_pos =
    assert(pos >= 0 && end_pos >= 0 &&
	     pos <= !buf_dim && end_pos <= !buf_dim);
    Platform.find_lf !bufs pos end_pos in

object(self)

  method private lazy_init() =
    Plasma_client.sync self#lazy_init_e ()

  method private lazy_init_e() =
    match !lazy_init_started with
      | None ->
	  let init_e =
	    self # buffer_init_e()
	    ++ (fun () -> self # fill_buffers_e index true)
	    ++ (fun () ->
		  self # cursor_init();
		  Netlog.logf `Debug
		    "Mapred_io.read_file(%s): init done" name;
		  lazy_init_started := None;
		  lazy_init := true;
		  eps_e (`Done()) esys
	       ) in
	  lazy_init_started := Some init_e;
	  init_e
      | Some init_e ->
	  init_e

  method private buffer_init_e() =
    Plasma_client.blocksize_e c
    ++ (fun bs ->
	  bsize := bs;
	  let bsizeL = Int64.of_int bs in
	  bigblocksize := round_up rc#bigblock_size bs;
	  let bigblockmult = !bigblocksize / bs in
	  let bigblockmultL = Int64.of_int bigblockmult in
	  (* bigblockmult: how many blocks fit into a bigblock *)

	  let mr_buffer_size = round_up rc#mr_buffer_size bs in
	  let mr_buffer_size_tight = round_up rc#mr_buffer_size_tight bs in
	  let () =
	    if !bigblocksize > mr_buffer_size then
	      failwith
		"Mapred_io.read_file: bigblock bigger than mr_buffer_size" in
	  let () =
	    if !bigblocksize > mr_buffer_size_tight then
	      failwith
		"Mapred_io.read_file: bigblock bigger than mr_buffer_size_tight" in
	  let shm_mng = Plasma_client.shm_manager c in
	  let shm_cond = shm_mng # shm_condition in
	  let shm_id = shm_mng # new_id() in
	  let mem_limit = 
	    match shm_cond with
	      | `Normal -> mr_buffer_size
	      | `Tight -> mr_buffer_size_tight in
	  shm_undo := (fun () -> shm_mng # book_allocation shm_id 0);
	  Plasma_client.retry_e
	    c name
	    (fun () ->
	       Plasma_client.with_trans_e c 
		 (fun trans ->
		    Plasma_client.lookup_e trans name false
		    ++ (fun l_inode ->
			  inode := l_inode;
			  Plasma_client.get_inodeinfo_e trans l_inode
			  ++ (fun l_ii ->
				ii := l_ii;
				eps_e (`Done()) esys
			     )
		       )
		 )
	    ) ()
	  ++ (fun () ->
		(* Plasma_client.abort_cluster c; - impossible in async ctx *)
		let max_index = 
		  if !ii.eof = 0L then
		    (-1L) 
		  else
		    Int64.div (Int64.pred !ii.eof) bsizeL in
		let eof_index =
		  Int64.succ max_index in
		(** [max_index] is the index of the last block of the file.
		    [eof_index] is the following block.
		 *)
		Netlog.logf `Debug
		  "Mapred_io.read_file(%s,index=%Ld,len=%Ld,eofidx=%Ld)" 
		  name index len eof_index;
		if index > Int64.sub eof_index len then
		  failwith "Mapred_io.read_file: cannot read beyond eof";
		if Int64.rem index bigblockmultL <> 0L then
		  failwith 
		    "Mapred_io.read_file: index is not a multiple of bigblocks";
		if (Int64.rem len bigblockmultL <> 0L && 
		      index < Int64.sub eof_index len
		         (* = not reading until eof *) ) then
		  failwith
		    "Mapred_io.read_file: len is not a multiple of bigblocks";

		end_index := Int64.add index len;
		hard_end_index := 
		  min (Int64.add !end_index bigblockmultL) eof_index;
		Netlog.logf `Debug
		  "Mapred_io.read_file(%s,index=%Ld,len=%Ld,endidx=%Ld,hendidx=%Ld)" 
		  name index len !end_index !hard_end_index;
		let buf_limit = max bigblockmult (mem_limit / bs) in
		let buf_numL = (* min hard_len *) (Int64.of_int buf_limit) in
		buf_num := Int64.to_int buf_numL;
		buf_dim := !buf_num * bs;
		let (l_buf_fd,buf_name) = Plasma_util.get_shm_fd "mapred" 0 in
		buf_fd := l_buf_fd;
		bufs :=
		  ( try
		      Netsys_posix.shm_unlink buf_name;
		      Netlog.logf `Debug 
			"read_file (pid %d): truncate fd=%Ld eof=%d"
			(Unix.getpid()) 
			(Netsys.int64_of_file_descr l_buf_fd) !buf_dim;
		      Plasma_util.enlarge_shm l_buf_fd !buf_dim;
		      shm_mng # book_allocation shm_id !buf_dim;
		      Plasma_util.map_shm l_buf_fd 0L !buf_dim
		    with
		      | error -> Unix.close l_buf_fd; raise error
		  );
		eps_e (`Done ()) esys
	     )
       )
    

  method private cursor_init() =
    (** Unless we are reading from the beginning of the file, the first
	line to read starts after the first LF.
     *)
    if index > 0L then (
      try
	let p = find_lf 0 !buf_end in
	if p >= !bigblocksize then raise Not_found;
	buf_cursor := p+1
      with
	| Not_found ->
	    failwith "Mapred_io.read_file: record too large"
    )


  method private end_of_file() =
    Netlog.logf `Debug
      "Mapred_io.read_file(%s,EOF,pos_in=%d)" name !pos_in;
    raise End_of_file

  method pos_in = !pos_in

  method peek_record() =
    if !closed then failwith "Mapred_io.read_file: closed";
    if !to_fd_engine <> None then 
      failwith "Mapred_io.read_file: The fd engine is still running";
    if not !lazy_init then self#lazy_init();
    match !cur_rec with
      | None -> 
	  let line = self # input_record_rest "" in
	  if line <> "" && line.[0] = '\000' then
	    Netlog.logf `Warning
	      "Mapred_io.read_file: Line starts with NUL byte pos_in=%d" 
	      !pos_in;
	  cur_rec := Some line;
	  line
      | Some r -> r

  method input_record() =
    if !closed then failwith "Mapred_io.read_file: closed";
    if !to_fd_engine <> None then 
      failwith "Mapred_io.read_file: The fd engine is still running";
    if not !lazy_init then self#lazy_init();
    match !cur_rec with
      | None -> 
	  let line = self # input_record_rest "" in
	  if line <> "" && line.[0] = '\000' then
	    Netlog.logf `Warning
	      "Mapred_io.read_file: Line starts with NUL byte pos_in=%d"
	      !pos_in;
	  cur_rec := None;
	  incr pos_in;
	  line
      | Some r -> 
	  cur_rec := None;
	  incr pos_in;
	  r

  (** Read the remaining part of the current line. The first part of the
      line is [line0]. The first part can be passed as [""] which means
      that nothing is read yet. The method raises [End_of_file] if 
      [line0=""] and there is no more line.
   *)
  method private input_record_rest line0 =
    if !buf_eof_reached then self # end_of_file();
    assert(!lazy_init && !buf_cursor <= !buf_end);
    (** Note that [buf_cursor=buf_end] is possible when the previous line
	ended at the last byte of the buffer. In this case we always run
	into the [Not_found] handler below, and this line can only be
	read after a refill. (Another cause is a totally empty file.)
     *)
    let l0 = String.length line0 in
    try
      let p = find_lf !buf_cursor !buf_end in
      let l = l0 + p - !buf_cursor in
      if l >= !bigblocksize then
	failwith "Mapred_io.read_file: record too large";
      let line = String.create l in
      if l0 > 0 then
	String.blit line0 0 line 0 l0;
      Netsys_mem.blit_memory_to_string !bufs !buf_cursor line l0 (l-l0);
      buf_cursor := p+1;
      (** If we know already where the requested region ends, we can check
	  whether this is the last line. [p] is the position of the LF
	  char of the current line. If the LF byte is in the overflow
	  region, this is the last line. (Note that [p=eof-1] just means
	  that the last line is the following line which is completely
	  taken from the overflow region.)
       *)
      ( match !buf_eof with
	  | None -> ()
	  | Some eof ->
	      if p >= eof then
		buf_eof_reached := true
      );
      line
    with
      | Not_found ->
	  (** The LF char cannot be found. If this is already the
	      inner call [input_record_rest], i.e. [line0 <> ""],
	      there is no chance to extract a line of at most the
	      size of bigblocks. So fail. If [line0 = ""], we extract
	      here the new version of [line0] for the inner call.
	   *)
	  if l0 > 0 || !buf_end - !buf_cursor >= !bigblocksize then
	    failwith "Mapred_io.read_file: record too large";
	  let l0 = !buf_end - !buf_cursor in
	  let line0 = String.create l0 in
	  Netsys_mem.blit_memory_to_string !bufs !buf_cursor line0 0 l0;
	  (** We refill here. We do this even if [buf_eof] is already set.
	      When [buf_eof] is set, we probably already have read a part
	      of the overflow region. Nevertheless, we need to ensure that
	      we read at least a whole bigblock from the overflow region,
	      and the remaining bigblock might become available only after
	      a refill. Note that the refill might lead to empty buffers
	      when we are at the EOF position of the file.
	   *)
	  self # fill_buffers 
	    (Int64.add !buf_index (Int64.of_int !buf_count));
	  if !buf_count = 0 then (
	    (** No more data exist. This means the current line is the
		last line, and it does not end with an LF char.
	     *)
	    buf_eof_reached := true;
	    (** [l0=0] can happen when the length of the file is
		a whole multiple of bigblocks. The previous line ended
		at the end of the previous bigblock, and there are no
		more lines to read.
	     *)
	    if l0 = 0 then self # end_of_file();
	    line0
	  ) else
	    (** Read the remaining part of the line *)
	    self # input_record_rest line0
		    

  method to_fd_e fd =
    (** We write everything until the end of the requested region, 
	plus the beginning of the overflow region until the first
	LF (or EOF).
     *)
    if !closed then failwith "Mapred_io.read_file: closed";
    if !to_fd_engine <> None then 
      failwith "Mapred_io.read_file: The fd engine is still running";
    let io_dev =
      `Polldescr(Netsys.get_fd_style fd,fd,esys) in

    let rec out_loop() =
      assert (!lazy_init);
      match !buf_eof with
	| None ->
	    (* Write whole buffer *)
	    out_until !buf_end ++ out_loop
	| Some eof ->
	    (* If the cursor is before eof, just write until eof *)
	    if !buf_cursor < eof then
	      out_until eof ++ out_loop
	    else (
	      (* Search for the LF byte in the overflow region *)
	      try
		let p = find_lf !buf_cursor !buf_end in
		if p - !buf_cursor >= !bigblocksize then
		  record_too_large()
		else
		  out_until (p+1) ++ out_eof
	      with
		| Not_found ->
		    let l0 = !buf_end - !buf_cursor in
		    out_until !buf_end
		    ++ (fun () ->
			  try
			    let q = find_lf !buf_cursor !buf_end in
			    if l0 + q - !buf_cursor >= !bigblocksize then
			      record_too_large()
			    else
			      out_until (q+1) ++ out_eof
			  with
			    | Not_found ->
				(* The LF of the last line might be missing *)
				let l = l0 + !buf_end - !buf_cursor in
				if l >= !bigblocksize 
				then
				  record_too_large()
				else
				  let lf_flag = l > 0 in
				  out_until !buf_end ++ out_lf lf_flag 
				  ++ out_eof
		       )
	    )
	      
    and out_until p =
      (* Output until byte pos p *)
      Netlog.logf `Debug
	"Mapred_io.read_file(%s): out_until start=%d end=%d" name !buf_cursor p;
      Uq_io.really_output_e
	io_dev
	(`Memory !bufs)
	!buf_cursor
	(p - !buf_cursor)
      ++ (fun () ->
	    Netlog.logf `Debug
	      "Mapred_io.read_file(%s): out_until done" name;
	    buf_cursor := p;
	    if p = !buf_end then
	      self # fill_buffers_e 
		(Int64.add !buf_index (Int64.of_int !buf_count)) false
	    else
	      eps_e (`Done()) esys
	 )
	
    and out_lf cond () =
      (* Output a single LF char if conf *)
      Netlog.logf `Debug
	"Mapred_io.read_file(%s): out_lf cond=%B" name cond;
      if cond then
	Uq_io.output_string_e io_dev "\n"
      else
	eps_e (`Done ()) esys

    and out_eof() =
      Netlog.logf `Debug
	"Mapred_io.read_file(%s): out_eof" name;
      buf_eof_reached := true;
      eps_e (`Done ()) esys

    and record_too_large () =
      eps_e (`Error (Failure "Mapred_io.read_file: record too large")) esys

    in
    let e = 
      if !lazy_init then
	out_loop()
      else
	self#lazy_init_e() ++ out_loop in
    to_fd_engine := Some e;
    let forget _ = to_fd_engine := None in
    Uq_engines.when_state ~is_done:forget ~is_error:forget ~is_aborted:forget e;
    e

  method private fill_buffers i =
    Plasma_client.sync (self#fill_buffers_e i) false 

  method private fill_buffers_e i fakeinit =
    (** Fill buffers so that the first one has index i *)
    
    assert(fakeinit || !lazy_init);

    (** [len_blocks] is the number of blocks we read here, either until
	the buffers are filled, or the end of the overflow region is
	reached.
     *)

    let buf_numL = Int64.of_int !buf_num in
    let bsizeL = Int64.of_int !bsize in

    let len_blocks = 
      Int64.to_int (min (Int64.sub !hard_end_index i) buf_numL) in
    let len = len_blocks * !bsize in  (* rounded up to blocks *)
    let lenL = Int64.of_int len in
    let pos = Int64.mul i bsizeL in

    Netlog.logf `Debug
      "Mapred_io.read_file: copy_out inode=%Ld pos=%Ld len=%d"
      !inode pos len;

    Plasma_client.retry_e
      c name
      (fun () ->
	 ignore(Unix.lseek !buf_fd 0 (Unix.SEEK_SET));
	 Plasma_client.copy_out_e c !inode pos !buf_fd lenL
      )
      ()
    ++ (fun () ->
	  let n = Int64.to_int (max (min lenL (Int64.sub !ii.eof pos)) 0L) in
	  (** This are the number of bytes we've really read into the buffer.
	      Note that [pos > ii.eof] is possible.
	   *)

	  let log_lenL = Int64.mul (Int64.sub !end_index i) bsizeL in
	  let m = 
	    if log_lenL < 0L || pos > !ii.eof then 
	      0
	    else
	      Int64.to_int (min log_lenL (Int64.sub !ii.eof pos)) in
	  (** This is the byte position of the end of the requested region
	      relative to [pos]. Negative values would mean that this end
	      is already behind us - we handle this as 0 instead.
	      
	      If [m > n] the end of the requested region is not yet reached.
	      If [m = n] the end of the requested region is at the end of the
	      buffer. If [m < n] the end is before the buffer or in the buffer.
	   *)

	  (* copy_out does not look at eof, but dumbly tries to get the
	     blocks
	   *)
	  let eof = Int64.add pos lenL >= !ii.eof in

	  buf_index := i;
	  buf_end := n;
	  buf_count := if len=0 then 0 else (len - 1) / !bsize + 1;
	  buf_eof := 
	    if eof || Int64.add i (Int64.of_int !buf_count) >= !end_index then
	      (* return the logical EOF here, not the hard EOF *)
	      Some m
	    else
	      None;
	  buf_cursor := 0;

	  Netlog.logf `Debug
	    "Mapred_io.read_file: new buffer: index=%Ld end=%d count=%d eof=%s"
	    !buf_index !buf_end !buf_count
	    (match !buf_eof with None -> "none" | Some n -> string_of_int n);
	  
	  (* Plasma_client.abort_cluster c; - impossible in async ctx *)
	  eps_e (`Done ()) esys
       )

  method close_in() =
    if !closed then (
      ( match !to_fd_engine with
	  | None -> ()
	  | Some e -> e # abort(); to_fd_engine := None
      );
      Plasma_client.abort_cluster c;  (* CHECK *)
      Unix.ftruncate !buf_fd 0;
      Unix.close !buf_fd;
      Netsys_mem.memory_unmap_file !bufs;
      !shm_undo();
      closed := true
    )

  method abort() =
    self # close_in()

  method plasma_cluster =
    c

end

type sync_readers =
    (unit -> record_reader) list

type async_readers =
    (unit -> record_reader Uq_engines.engine) list


let read_multiple c rc
                  ~(readers :[`Sync of sync_readers | `Async of async_readers])
                  () =
  let esys = Plasma_client.event_system c in
  let readers = ref readers in
  let pos_in = ref 0 in
  let cur_rec = ref None in
  let cur_rd = ref None in


  let rec next_record() =
    match !cur_rd with
      | Some rd ->
	  ( try
	      rd # input_record()
	    with
	      | End_of_file ->
		  rd # close_in();
		  cur_rd := None;
		  next_reader()
	  )
      | None ->
	 next_reader()

  and next_reader() =
    match !readers with
      | `Sync(first :: rest) ->
	  readers := `Sync rest;
	  let rd = first() in
	  cur_rd := Some rd;
	  next_record()
      | `Async(first :: rest) ->
	  readers := `Async rest;
	  let rd = Plasma_client.sync first () in
	  cur_rd := Some rd;
	  next_record()
      | `Sync [] | `Async [] ->
	  raise End_of_file
  in

object(self)
  method pos_in = !pos_in

  method peek_record() =
    match !cur_rec with
      | None -> 
	  let line = next_record() in
	  cur_rec := Some line;
	  line
      | Some r -> r

  method input_record() =
    match !cur_rec with
      | None ->
	  let line = next_record() in
	  incr pos_in;
	  line
      | Some r ->
	  cur_rec := None;
	  incr pos_in;
	  r

  method to_fd_e fd =
    let io_dev =
      `Polldescr(Netsys.get_fd_style fd,fd,esys) in

    Netlog.logf `Debug
      "Mapred_io.read_multiple: to_fd_e start";
    let rec out_loop() =
      match !cur_rec with
	| Some line ->
	    cur_rec := None;
	    Uq_io.output_string_e io_dev line
	    ++ (fun () -> Uq_io.output_string_e io_dev "\n")
	    ++ out_loop
	| None ->
	    ( match !cur_rd with
		| Some rd ->
		    cur_rd := None;
		    rd # to_fd_e fd
		    ++ (fun () -> rd # close_in(); out_loop())
		| None ->
		    ( match !readers with
			| `Async(first :: rest) ->
			    readers := `Async rest;
			    first()
			    ++ (fun rd ->
				  cur_rd := Some rd;
				  out_loop()
			       )
			| `Sync(first :: rest) ->
			    readers := `Sync rest;
			    cur_rd := Some(first());
			    out_loop()
			| `Async [] | `Sync [] ->
			    Netlog.logf `Debug
			      "Mapred_io.read_multiple: to_fd_e end";
			    eps_e (`Done()) esys
		    )
	    )
    in
    out_loop()

  method close_in() =
    match !cur_rd with
      | Some rd ->
	  rd # close_in()
      | None ->
	  ()

  method abort() =
    match !cur_rd with
      | Some rd ->
	  rd # abort()
      | None ->
	  ()

  method plasma_cluster =
    c
end


let x_write_file c rc name =
  let esys = Plasma_client.event_system c in

  (* This block of variables is initialized in buf_init_e: *)
  let bsize = ref 0 in
  let bigblocksize = ref 0 in
  let inode = ref 0L in
  let buf_len = ref 0 in        (* length of bufs in bytes *)
  let bufs = ref zero_mem in    (* The buffer as bigarray *)
  let buf_fd = ref Unix.stdin in  (* The buffer as descriptor *)
  let shm_undo = ref (fun () -> ()) in    (* deregisters the buffer in shm *)

  (* The following variables are created with their right initial value: *)
  let buf_end = ref 0 in        (* end position in bufs in bytes *)
  let buf_pos = ref 0L in       (* byte position in file of first buf byte *)

  let lazy_init = ref false in        (* whether lazy_init_e is done *)
  let lazy_init_started = ref None in
  let pos_out = ref 0 in
  let bpos_out = ref 0L in
  let closed = ref false in
  let from_fd_engine = ref None in

  let rev_find_lf pos end_pos =
    assert(!lazy_init && pos >= 0 && end_pos >= 0 &&
	     pos <= !buf_len && end_pos <= !buf_len);
    Platform.rev_find_lf !bufs pos end_pos
  in 

object(self)

  method private lazy_init() =
    Plasma_client.sync self#lazy_init_e ()

  method private lazy_init_e() =
    match !lazy_init_started with
      | None ->
	  let init_e = self # buffer_init_e() in
	  lazy_init_started := Some init_e;
	  init_e
      | Some init_e ->
	  init_e

  method private buffer_init_e() =
    Plasma_client.blocksize_e c
    ++ (fun bs ->
	  bsize := bs;
	  bigblocksize := round_up rc#bigblock_size bs;
	  let mr_buffer_size = round_up rc#mr_buffer_size bs in
	  let mr_buffer_size_tight = round_up rc#mr_buffer_size_tight bs in
	  if !bigblocksize > mr_buffer_size then
	    failwith
	      "Mapred_io.write_file: bigblock bigger than mr_buffer_size";
	  if !bigblocksize > mr_buffer_size_tight then
	    failwith 
	      "Mapred_io.write_file: bigblock bigger than mr_buffer_size_tight";
	  let shm_mng = Plasma_client.shm_manager c in
	  let shm_cond = shm_mng # shm_condition in
	  let shm_id = shm_mng # new_id() in
	  let mem_limit = 
	    match shm_cond with
	      | `Normal -> mr_buffer_size
	      | `Tight -> mr_buffer_size_tight in
	  shm_undo := (fun () -> shm_mng # book_allocation shm_id 0);
	  
	  Plasma_client.retry_e
	    c name
	    (fun () ->
	       Plasma_client.with_trans_e c
		 (fun trans ->
		    Plasma_client.lookup_e trans name false
		    ++ (fun l_inode ->
			  inode := l_inode;
			  Plasma_client.get_inodeinfo_e trans l_inode
			  ++ (fun ii ->
				let ii' = { ii with eof = 0L } in
				Plasma_client.set_inodeinfo_e trans l_inode ii'
			     )
		       )
		 )
	    )
	    ()
	  ++ (fun () ->
		(* Plasma_client.abort_cluster c; - impossible in async ctx *)
		Netlog.logf `Debug "Mapred_io.write_file(%s)" name;
		let buf_num = max 1 (mem_limit / bs) in
		(* let buf_numL = Int64.of_int buf_num in *)
		buf_len := buf_num * bs;
		let (l_buf_fd,buf_name) = Plasma_util.get_shm_fd "mapred" 0 in
		buf_fd := l_buf_fd;
		bufs :=
		  ( try
		      Netsys_posix.shm_unlink buf_name;
		      Netlog.logf `Debug
			"write_file (pid %d): truncate fd=%Ld eof=%d"
			(Unix.getpid()) (Netsys.int64_of_file_descr l_buf_fd) 
			!buf_len;
		      Plasma_util.enlarge_shm l_buf_fd !buf_len;
		      shm_mng # book_allocation shm_id !buf_len;
		      Plasma_util.map_shm l_buf_fd 0L !buf_len
		    with
		      | error -> Unix.close l_buf_fd; raise error
		  );
		lazy_init := true;
		eps_e (`Done()) esys
	     )
       )


  method pos_out = !pos_out
  method bpos_out = !bpos_out

  method output_record line =
    (* we do not protect against having LF in line. User's responsibility *)
    if !closed then failwith "Mapred_io.write_file: closed";
    if !from_fd_engine <> None then
      failwith "Mapred_io.write_file: The fd engine is still running";
    if not !lazy_init then self # lazy_init();
    assert (!lazy_init);
    if line <> "" && line.[0] = '\000' then
      Netlog.logf `Warning
	"Mapred_io.write_file: Line starts with NUL byte pos_out=%d" 
	!pos_out;
    let l = String.length line in
    if l >= !bigblocksize then
      failwith "Mapred_io.write_file: record too large";
    let p = ref 0 in
    while !p < l+1 do
      let space = !buf_len - !buf_end in
      let n =
	if !p < l then (
	  let k = min (l - !p) space in
	  Netsys_mem.blit_string_to_memory line !p !bufs !buf_end k;
	  k
	) else (
	  !bufs.{ !buf_end } <- '\n';
	  1
	) in
      buf_end := !buf_end + n;
      p := !p + n;
      if !buf_end = !buf_len then
	self # flush()
    done;
    bpos_out := Int64.add !bpos_out (Int64.of_int (l + 1))

  method flush() =
    if !closed then failwith "Mapred_io.write_file: closed";
    if !from_fd_engine <> None then
      failwith "Mapred_io.write_file: The fd engine is still running";
    if not !lazy_init then self # lazy_init();
    Netlog.logf `Debug "Mapred_io.write_file.flush: name=%s pos=%Ld len=%d"
      name !buf_pos !buf_end;
    Plasma_client.sync self#flush_e ()

  method private flush_e() =
    assert(!lazy_init);
    Plasma_client.retry_e
      c name
      (fun () ->
	 ignore(Unix.lseek !buf_fd 0 (Unix.SEEK_SET));
	 Plasma_client.copy_in_e
	   c !inode !buf_pos !buf_fd (Int64.of_int !buf_end) `Star
      )
      ()
    ++ (fun () ->
	  (* copy_in works only with multiples of bsize. So if we did not match
	     this, retain the end of the buffer. (The requirement here is
	     that !buf_pos must remain a multiple of bsize.)
	   *)
	  let tail = !buf_end mod !bsize in
	  if tail > 0 then (
	    let n = !buf_end - tail in
	    Bigarray.Array1.blit
	      (Bigarray.Array1.sub !bufs (!buf_end - tail) tail)
	      (Bigarray.Array1.sub !bufs 0 tail);
	    buf_pos := Int64.add !buf_pos (Int64.of_int n);
	    buf_end := tail
	  )
	  else (
	    buf_pos := Int64.add !buf_pos (Int64.of_int !buf_end);
	    buf_end := 0;
	  );
	  
	  (* Plasma_client.abort_cluster c; - impossible in async ctx *)
	  eps_e (`Done()) esys
       )

  method from_fd_e fd =
    if !from_fd_engine <> None then
      failwith "Mapred_io.write_file: The fd engine is still running";

    let io_dev =
      `Polldescr(Netsys.get_fd_style fd,fd,esys) in
    let ends_with_lf = ref true in

    let rec in_loop () =
      assert (!lazy_init);
      if !buf_end = !buf_len then
	self # flush_e () ++ in_loop
      else
	Uq_engines.meta_engine
	  (Uq_io.input_e io_dev (`Memory !bufs) !buf_end (!buf_len - !buf_end))
	++ (fun st ->
	      match st with
		| `Done n ->
		    if n > 0 then
		      ends_with_lf := !bufs.{ !buf_end + n - 1 } = '\n';
		    buf_end := !buf_end + n;
		    bpos_out := Int64.add !bpos_out (Int64.of_int n);
		    in_loop()
		| `Error End_of_file -> 
		    if not !ends_with_lf then (
		      (* add LF *)
		      !bufs.{ !buf_end } <- '\n';
		      incr buf_end;
	              bpos_out := Int64.add !bpos_out 1L
		    );
		    eps_e (`Done()) esys
		| `Error error -> eps_e (`Error error) esys
		| `Aborted -> eps_e `Aborted esys
	   )
    in
    let e =
      if !lazy_init then 
	in_loop()
      else
	self#lazy_init_e() ++ in_loop in
    from_fd_engine := Some e;
    let forget _ = from_fd_engine := None in
    Uq_engines.when_state ~is_done:forget ~is_error:forget ~is_aborted:forget e;
    e

    
  method from_fd_lim_e prepend fd lim : (string * bool) Uq_engines.engine =
    (** This method is not exported to the publc. It is ensured that at most
	lim bytes are finally output, i.e. finally we have [bpos_out <= lim].
	First the data is taken from the string [prepend] (which must fit
	into the buffer), and then data is read from the file descriptor [fd].
	The number of bytes read from [fd] is limited to
	[lim-bpos_out-String.length prepend] (when [bpos_out] is the current
	value). If the
	final line of this byte range is not ended (with LF), the
	remaining bytes are not output, but returned to the caller as string.
	The bool says whether we hit EOF.

	The buffer is always flushed at the end.
     *)
    if !from_fd_engine <> None then
      failwith "Mapred_io.write_file: The fd engine is still running";

    Netlog.logf `Debug
      "Mapred_io.write_file(%s): from_fd_lim_e starting" name;

    let io_dev =
      `Polldescr(Netsys.get_fd_style fd,fd,esys) in
    let ends_with_lf = ref true in
    let buf_linestart = ref !buf_end in
    (* buf_linestart is the buffer position where the last line of the
       buffer starts.
     *)

    let rec in_loop () =
      assert (!lazy_init);
      Netlog.logf `Debug
	"Mapred_io.write_file(%s): in_loop buf_linestart=%d buf_end=%d"
	name !buf_linestart !buf_end;
      if !buf_end = !buf_len then (
	(* We cannot output the last unterminated line of the buffer.
	   So this data is removed here, the buffer is flushed, and the
	   line is added again. It is theoretically possible that nothing
	   could be flushed - in this case we raise an exception.
	   This could be implemented a bit more clever without extra
	   string (flush does not destroy the buffer contents).
	 *)
	let line_len = !buf_end - !buf_linestart in
	let line = String.create line_len in
	Netsys_mem.blit_memory_to_string !bufs !buf_linestart line 0 line_len;
	buf_end := !buf_linestart;
	let old_buf_end = !buf_end in
	self # flush_e ()
	++ (fun () ->
	      assert (!buf_end <= old_buf_end);
	      if !buf_end = old_buf_end then
		eps_e
		  (`Error (Failure "Mapred_io.write_file: record too large"))
		  esys
	      else (
		buf_linestart := !buf_end;
		Netsys_mem.blit_string_to_memory line 0 !bufs !buf_end line_len;
		buf_end := !buf_end + line_len;
		in_loop()
	      )
	   )
      )
      else (
	let total_len = lim - Int64.to_int !bpos_out in
	if total_len = 0 then (
	  (* This means we reached the limit. *)
	  Netlog.logf `Debug
	    "Mapred_io.write_file(%s): in_loop hitting limit"
	    name;
	  let line_len = !buf_end - !buf_linestart in
	  let line = String.create line_len in
	  Netsys_mem.blit_memory_to_string !bufs !buf_linestart line 0 line_len;
	  buf_end := !buf_linestart;
	  self # flush_e() 
	  ++ (fun () -> eps_e (`Done (line, false)) esys)
	)
	else (
	  let len = min (!buf_len - !buf_end) total_len in
	  Netlog.logf `Debug
	    "Mapred_io.write_file(%s): in_loop input len=%d"
	    name len;
	  Uq_engines.meta_engine
	    (Uq_io.input_e io_dev (`Memory !bufs) !buf_end len)
	  ++ (fun st ->
		match st with
		  | `Done n ->
		      Netlog.logf `Debug
			"Mapred_io.write_file(%s): in_loop input got %d"
			name n;
		      if n > 0 then (
			ends_with_lf := !bufs.{ !buf_end + n - 1 } = '\n';
			try
			  let p = 
			    rev_find_lf !buf_linestart (!buf_end+n) in
			  buf_linestart := p+1
			with
			  | Not_found -> ()
		      );
		      buf_end := !buf_end + n;
		      bpos_out := Int64.add !bpos_out (Int64.of_int n);
		      in_loop()
		  | `Error End_of_file -> 
		      Netlog.logf `Debug
			"Mapred_io.write_file(%s): in_loop input got EOF"
			name;
		      if not !ends_with_lf then (
			(* add LF *)
			!bufs.{ !buf_end } <- '\n';
			incr buf_end;
			bpos_out := Int64.add !bpos_out 1L
		      );
		      self # flush_e ()
		      ++ (fun () -> eps_e (`Done ("", true)) esys)
		  | `Error error -> eps_e (`Error error) esys
		  | `Aborted -> eps_e `Aborted esys
	     )
	)
      )
    in
    let do_prepend() =
      let plen = String.length prepend in
      if !buf_len - !buf_end >= plen then (
	Netsys_mem.blit_string_to_memory prepend 0 !bufs !buf_end plen;
	buf_end := !buf_end + plen;
	eps_e (`Done()) esys
      )
      else (
	(* All we can do is to flush once. If this does not make enough
	   space. [prepend] is too long
	 *)
	self # flush_e() 
	++ (fun () ->
	      if !buf_len - !buf_end >= plen then (
		Netsys_mem.blit_string_to_memory prepend 0 !bufs !buf_end plen;
		buf_end := !buf_end + plen;
		eps_e (`Done()) esys
	      )
	      else
		eps_e 
		  (`Error (Failure "Mapred_io.write_file: record too large"))
		  esys
	   )
      ) in
    let e =
      if !lazy_init then 
	do_prepend() ++ in_loop
      else
	self#lazy_init_e() ++ do_prepend ++ in_loop in
    from_fd_engine := Some (e ++ (fun _ -> eps_e (`Done ()) esys));
    let forget _ = from_fd_engine := None in
    Uq_engines.when_state ~is_done:forget ~is_error:forget ~is_aborted:forget e;
    e

    
  method close_out() =
    if not !closed then (
      self # flush();
      Unix.ftruncate !buf_fd 0;
      Unix.close !buf_fd;
      Netsys_mem.memory_unmap_file !bufs;
      !shm_undo();
      closed := true
    )

  method abort() =
    if not !closed then (
      (* no flush here *)
      ( match !from_fd_engine with
	  | None -> ()
	  | Some e -> e # abort(); from_fd_engine := None
      );
      Unix.close !buf_fd;
      Netsys_mem.memory_unmap_file !bufs;
      !shm_undo();
      closed := true
    )

  method plasma_cluster =
    c
end


let write_file c rc name =
  (x_write_file c rc name :> record_writer)


let write_multiple c rc prefix limit ~create_sync ?create_async () =
  let esys = Plasma_client.event_system c in
  let n = ref 0 in
  let current = ref None in
  let pos_out = ref 0 in
  let bpos_out = ref 0L in

  let switch() = 
    Netlog.logf `Debug
      "Mapred_io.write_multiple(%s): switch n=%d" prefix !n;
    let name = create_sync prefix !n in
    incr n;
    Netlog.logf `Debug
      "Mapred_io.write_multiple(%s): switch sub-writer: %s" prefix name;
    let w = x_write_file c rc name in
    current := Some w;
    w in

  let switch_e() =
    match create_async with
      | None -> 
	  eps_e (`Done (switch())) esys
      | Some create_e ->
	  Netlog.logf `Debug
	    "Mapred_io.write_multiple(%s): switch n=%d" prefix !n;
	  let k = !n in
	  incr n;
	  create_e prefix k
	  ++ (fun name ->
		Netlog.logf `Debug
		  "Mapred_io.write_multiple(%s): switch sub-writer: %s"
		  prefix name;
		let w = x_write_file c rc name in
		current := Some w;
		eps_e (`Done w) esys
	     ) in
object(self)
  method pos_out = !pos_out
  method bpos_out = !bpos_out

  method output_record line =
    let w =
      match !current with
	| None ->
	    switch()
	| Some w ->
	    let p =
	      Int64.add w#bpos_out (Int64.of_int (String.length line + 1))  in
	    if p >= limit then (
	      w # close_out();
	      switch()
	    )
	    else
	      w in
    w # output_record line

  method from_fd_e fd =
    Netlog.logf `Debug
      "Mapred_io.write_multiple(%s): from_fd_e starting" prefix;
    let rec loop prepend w =
      Netlog.logf `Debug
	"Mapred_io.write_multiple(%s): from_fd_e loop #prepend=%d"
	prefix (String.length prepend);
      w # from_fd_lim_e prepend fd (Int64.to_int limit)
      ++ (fun (prepend',eof) ->
	    if eof then
	      eps_e (`Done ()) esys
	    else (
	      w # abort();  (* non-blocking. w is already flushed *)
	      switch_e()
	      ++ (fun w' -> loop prepend' w')
	    )	    
	 )
    in
    ( match !current with
	| None -> switch_e()
	| Some w -> eps_e (`Done w) esys
    )
    ++ (fun w -> loop "" w)

  method flush() =
    match !current with
      | None -> ()
      | Some w -> w # flush()

  method close_out () =
    match !current with
      | None -> ()
      | Some w -> w # close_out()

  method abort() =
    match !current with
      | None -> ()
      | Some w -> w # abort()

  method plasma_cluster =
    c
end


let create_file_e ?(repl=0) c name =
  let esys = Plasma_client.event_system c in
  Plasma_client.retry_e c name
    (fun () ->
       Plasma_client.with_trans_e c
	 (fun trans ->
	    let ii = Plasma_client.regular_ii c 0o666 in
	    let ii = { ii with replication = repl } in
	    Plasma_client.create_file_e trans name ii
	    ++ (fun _ -> 
		  Netlog.logf `Debug "create_file name=%s" name;
		  eps_e (`Done()) esys)
	 )
    )
    ()

let create_file ?repl c name =
  Plasma_client.sync (create_file_e ?repl c) name


let delete_file c name =
  Plasma_client.retry c name
    (fun () ->
       Plasma_client.with_trans c
	 (fun trans ->
	    let _ = Plasma_client.unlink trans name in
	    Netlog.logf `Debug "delete_file name=%s" name
	 )
    )
    ()


let file_blocks c name =
  Plasma_client.retry c name
    (fun () ->
       Plasma_client.with_trans c
	 (fun trans ->
	    let bsize = Plasma_client.blocksize c in
	    let bsizeL = Int64.of_int bsize in
	    let ii = 
	      Plasma_client.get_inodeinfo trans 
		(Plasma_client.lookup trans name false) in
	    let blocks =
	      if ii.eof = 0L then 0L 
	      else
		Int64.succ (Int64.div (Int64.pred ii.eof) bsizeL) in
	    blocks
	 )
    )
    ()


