(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_task_exec.ml 280 2010-10-27 18:13:07Z gerd $ *)

open Mapred_def
open Mapred_tasks
open Plasma_rpcapi_aux
open Printf
open Uq_engines.Operators

let run_stack f s =
  while not (Stack.is_empty s) do
    f (Stack.pop s)
  done


let fragment_reader c rc file =
  let esys = Plasma_client.event_system c in
  let file' =
    List.map
      (fun (filename,block,len) ->
	 let len' =
	   if len = 0L then
	     Mapred_io.file_blocks c filename
	   else
	     len in
	 (filename,block,len')
      )
      file in
  Mapred_io.read_multiple
    c rc
    ~readers:(`Async
	       (List.map
		  (fun (filename,block,len) ->
		     (fun () ->
			Netlog.logf `Debug
			  "fragment_reader(%s,%Ld,%Ld)" filename block len;
			let rd = Mapred_io.read_file c rc filename block len in
			eps_e (`Done rd) esys
		     )
		  )
		  file'
	       ))
    ()


let delete_files c file =
  let ht = Hashtbl.create 5 in
  List.iter (fun (name,_,_) -> Hashtbl.replace ht name ()) file;
  Hashtbl.iter
    (fun name () ->
       Mapred_io.delete_file c name
    )
    ht


let record_config me jc =
object
  method bigblock_size = jc#bigblock_size
  method mr_buffer_size = me#config#mr_buffer_size
  method mr_buffer_size_tight = me#config#mr_buffer_size_tight
end


let exec_map_task me jc mj (t:map_task) ti =
  let cleanup = Stack.create() in
  try
    let esys = Unixqueue.create_unix_event_system() in
    let c = me # open_cluster esys in
    Stack.push (fun () -> Plasma_client.close_cluster c) cleanup;

    (* The output of a map task should only be written to the local node.
       Also, be careful and set replication to 1.
     *)
    Plasma_client.configure_pref_nodes c (Plasma_client.local_identities c);


    let rc = record_config me jc in
    let reader = fragment_reader c rc t.map_input in
    Stack.push (fun () -> reader # close_in()) cleanup;
    
    let wfiles = ref [] in

    let create_async prefix k =
      let name = sprintf "%s%05d" prefix k in
      Mapred_io.create_file_e ~repl:1 c name
      ++ (fun () ->
	    (* Plasma_client.abort_cluster c; - impossible in async context *)
	    wfiles := [name, 0L, 0L] :: !wfiles;
	    eps_e (`Done name) esys
	 ) in
    let create_sync prefix k =
      Plasma_client.sync (create_async prefix) k in

    let writer =
      Mapred_io.write_multiple
	c rc t.map_output_prefix (Int64.of_int me#config#mr_sort_size)
	~create_sync ~create_async () in
    Stack.push (fun () -> writer # close_out()) cleanup;
    
    mj # map me jc ti reader writer;

    run_stack (fun f -> f()) cleanup;
    
    `Ok !wfiles
  with
    | error ->
	let msg =
	  sprintf "Exception %s%s"
	    (Netexn.to_string error)
	    (if Printexc.backtrace_status() then
	       " - backtrace: " ^ (Printexc.get_backtrace())
	     else ""
	    ) in
	run_stack (fun f -> try f() with _ -> ()) cleanup;
	`Error msg


let ba_substring ba pos len =
  let s = String.create len in
  Netsys_mem.blit_memory_to_string ba pos s 0 len;
  s
	
let exec_sort_task me jc mj (t:sort_task) =
  (* very dumb impl: read everything into an array, sort it, write it.

     ideas for improvements:
     - if the input file fits into the reader buffer, the strings
       could be just pointers into this buffer (avoid copying, or
       do just temporary copies that can be minor-collected)
   *)
  let cleanup = Stack.create() in
  try
    Netlog.logf `Debug "Start sort %s"
      (Mapred_tasks.string_of_task_id (`Sort t));


    let esys = Unixqueue.create_unix_event_system() in
    let c = me # open_cluster esys in
    Stack.push (fun () -> Plasma_client.close_cluster c) cleanup;

    (* The output of a sort task should only be written to the local node.
       Also, be careful and set replication to 1.
     *)
    Plasma_client.configure_pref_nodes c (Plasma_client.local_identities c);

    Mapred_io.create_file ~repl:1 c t.sort_output;
    
    let rc = record_config me jc in
    let reader = fragment_reader c rc t.sort_input in
    Stack.push (fun () -> reader # close_in()) cleanup;

    let sort_limit = me#config#mr_sort_size in

    let b_data = 
      Bigarray.Array1.create Bigarray.char Bigarray.c_layout sort_limit in
    let b_eof = ref 0 in
    let b_size = ref 1024 in
    let b_end = ref 0 in
    let b_lines  = ref (Array.make !b_size 0) in
    (* let b_keys   = ref (Array.make !b_size "") in *)
    let b_hashes = ref (Array.make !b_size 0) in

    let double a null new_size =
      let old_a = !a in
      a := Array.make new_size null;
      Array.blit old_a 0 !a 0 !b_size
    in

    let extract = mj # extract_key me jc in

    ( try
	while true do
	  let line = reader # input_record() in
	  let key  = extract line in
	  let hash = Hashtbl.hash key in
	  if !b_end = !b_size then (
	    let new_size = 2 * !b_size in
	    double b_lines 0 new_size;
	    (* double b_keys "" new_size; *)
	    double b_hashes 0 new_size;
	    b_size := new_size
	  );
	  let len = String.length line in
	  let old_eof = !b_eof in
	  let new_eof = !b_eof + len in
	  if new_eof > sort_limit then
	    failwith "Amount of data exceeds sort limit";
	  Netsys_mem.blit_string_to_memory line 0 b_data !b_eof len;
	  b_eof := !b_eof + len;
	  !b_lines. ( !b_end ) <- old_eof;
	  (* !b_keys.  ( !b_end ) <- key; *)
	  !b_hashes.( !b_end ) <- hash;
	  incr b_end
	done
      with
	| End_of_file -> ()
    );

    let index_array = Array.init !b_end (fun k -> k) in
    
    let get_line i =
      let i' = i+1 in
      let offs = !b_lines.(i) in
      let offs' = if i' = !b_end then !b_eof else !b_lines.(i') in
      ba_substring b_data offs (offs' - offs) 
    in

    Mapred_sort.merge_sort
      (fun i1 i2 ->
	 let h1 = Array.unsafe_get !b_hashes i1 in
	 let h2 = Array.unsafe_get !b_hashes i2 in
	 if h1 = h2 then
	   (*
	   let k1 = Array.unsafe_get !b_keys i1 in
	   let k2 = Array.unsafe_get !b_keys i2 in
	    *)
	   let l1 = get_line i1 in
	   let l2 = get_line i2 in
	   let k1 = extract l1 in
	   let k2 = extract l2 in
	   String.compare k1 k2
	 else
	   h1 - h2   (* h1, h2 are both positive *)
      )
      index_array;

    b_hashes := [| |];
    
    let writer =
      Mapred_io.write_file c rc t.sort_output in
    Stack.push (fun () -> writer # close_out()) cleanup;
    
    Array.iter
      (fun i ->
	 writer # output_record (get_line i)
      )
      index_array;
    writer # close_out();
    Netlog.logf `Debug "Done sort %s"
      (Mapred_tasks.string_of_task_id (`Sort t));

    reader # close_in();
    if t.sort_input_del then
      delete_files c t.sort_input;

    run_stack (fun f -> f()) cleanup;
    
    `Ok [[t.sort_output, 0L, 0L]]
  with
    | error ->
	let msg =
	  sprintf "Exception %s%s"
	    (Netexn.to_string error)
	    (if Printexc.backtrace_status() then
	       " - backtrace: " ^ (Printexc.get_backtrace())
	     else ""
	    ) in
	run_stack (fun f -> try f() with _ -> ()) cleanup;
	`Error msg


type shuffle_in =
    { in_index : int;   (* which input file *)
      in_key : string;  (* current key *)
      in_hash : int;    (* hash of the key *)
    }


module Rd = struct
  type t = shuffle_in

  (* This _must_ be the same criterion as for sorting! *)
  let compare (r1:t) (r2:t) =
    if r1.in_hash = r2.in_hash then (
      let d = String.compare r1.in_key r2.in_key in
      if d = 0 then
	r1.in_index - r2.in_index
      else
	d
    )
    else
      r1.in_hash - r2.in_hash
end

module RdSet = Set.Make(Rd)


let merging_reader input_array extract esys : Mapred_io.record_reader =
  let input_set = ref RdSet.empty in
  let () =
    for k = 0 to Array.length input_array - 1 do
      try
	let rd = input_array.(k) in
	let line = rd # peek_record() in  (* or End_of_file *)
	let key = extract line in
	let si =
	  { in_index = k;
	    in_key = key;
	    in_hash = Hashtbl.hash key
	  } in
	input_set := RdSet.add si !input_set
      with
	| End_of_file -> ()
    done in

  let pos_in = ref 0 in
  let current = ref None in

  let advance() =
    if !input_set = RdSet.empty then
      current := None
    else (
      let si = RdSet.min_elt !input_set in
      let rd = input_array.(si.in_index) in
      let key = si.in_key in
      let line = rd # input_record() in
      current := Some line;
      ( try
	  let next_line = rd # peek_record() in (* or End_of_file *)
	  let next_key = extract next_line in
	  if next_key <> key then (
	    input_set := RdSet.remove si !input_set;
	    let si' =
	      { si with 
		  in_key = next_key; 
		  in_hash = Hashtbl.hash next_key
	      } in
	    input_set := RdSet.add si' !input_set
	  )
	with
	  | End_of_file ->
	      input_set := RdSet.remove si !input_set
      )
    ) in

  let () = 
    advance() in

  ( object(self)
      method pos_in = !pos_in
      method input_record() =
	match !current with
	  | None -> raise End_of_file
	  | Some line -> 
	      advance();
	      line

      method peek_record() =
	match !current with
	  | None -> raise End_of_file
	  | Some line -> line
	      
      method close_in() =
	Array.iter (fun rd -> rd # close_in()) input_array

      method abort() =
	Array.iter (fun rd -> rd # abort()) input_array

      method to_fd_e fd = 
	let io_dev =
	  `Polldescr(Netsys.get_fd_style fd,fd,esys) in
	let input_many_records () =
	  let l = ref [] in
	  let n = ref 0 in
	  ( try
	      while !n < 1_000_000 do
		let line = self#input_record() in
		l := line :: !l;
		n := !n + String.length line + 1;
	      done;
	    with End_of_file -> ()
	  );
	  !n, String.concat "\n" (List.rev !l) in

	Netlog.logf `Debug
	  "Mapred_task_exec.exec_shuffle_task: reader to_fd_e";
	let stream =
	  Stream.from
	    (fun _ ->
	       let n, lines = input_many_records() in
	       Netlog.logf `Debug
		 "Mapred_task_exec.exec_shuffle_task: reader to_fd_e n=%d" n;
	       (* Unixqueue.Debug.enable := true; *)
	       if n > 0 then (
		 let e() = 
		   Uq_io.output_string_e io_dev lines
		   ++ (fun () -> Uq_io.output_string_e io_dev "\n") in
		 Some e
	       )
	       else (
		 Netlog.logf `Debug
		   "Mapred_task_exec.exec_shuffle_task: reader to_fd_e EOF";
		 None
	       )
	    ) in
	Uq_engines.stream_seq_engine () stream esys

      method plasma_cluster = input_array.(0) # plasma_cluster
    end
  )


let exec_shuffle_task me jc mj (t:shuffle_task) ti =
  let cleanup = Stack.create() in
  try
    let esys = Unixqueue.create_unix_event_system() in
    let c = me # open_cluster esys in
    Stack.push (fun () -> Plasma_client.close_cluster c) cleanup;

    (* In some situations we need a cluster that can always be used
       synchronously. We can set it to [c] when we know that [c] is
       never used asynchronously (or otherwise there is no conflicting use).
     *)
    let csync =
      if t.shuffle_reduce && List.length t.shuffle_input > 1 then (
	Netlog.logf `Debug
	  "Mapred_task_exec.exec_shuffle_task: using separate csync client";
	let c1 = me # open_cluster (Unixqueue.create_unix_event_system()) in
	Stack.push (fun () -> Plasma_client.close_cluster c1) cleanup;
	c1
      ) else
	(* if not t.shuffle_reduce: [reduce] is not called, so no danger.
	   if len t.shuffle_input = 1: No merge happens. Even if [reduce]
           is an asynchronous procedure, we know that there is no
	     synchronous line-by-line reading
	 *)
	c in

    (* Local node is always preferred *)
    Plasma_client.configure_pref_nodes c (Plasma_client.local_identities c);

    let rc = record_config me jc in
    let inputs =
      List.map
	(fun (file,_,_) ->
	   let reader = fragment_reader csync rc file in
	   Stack.push (fun () -> reader # close_in()) cleanup;
	   reader
	)
	t.shuffle_input in
    let input_array = Array.of_list inputs in

    let extract =
      mj # extract_key me jc in
    let part_key =
      mj # partition_of_key me jc in

    let input_files = Hashtbl.create 5 in
    List.iter2
      (fun (file,_,_) rd ->
	 let s =
	   String.concat "+"
	     (List.map 
		(fun (name,pos,len) -> sprintf "%s@%Ld/%Ld" name pos len)
		file) in
	 Hashtbl.add input_files rd ("[" ^ s ^ "]")
      )
      t.shuffle_input
      inputs;

    let (pmin,pmax) = t.shuffle_partitions in
    if pmin > pmax then failwith "Shuffle: empty set of output partitions";

    let output_part = 
      Array.make (pmax+1-pmin) None in
    
    let outputs =
      List.map
	(fun (filename,fpmin,fpmax) ->
	   if fpmin < pmin || fpmin > fpmax || fpmax > pmax then
	     failwith "Shuffle: bad output partitions";
	   let repl = if t.shuffle_reduce then 0 else 1 in
	   Mapred_io.create_file ~repl c filename;
	   let writer =
	     Mapred_io.write_file c rc filename in
	   Stack.push (fun () -> writer # close_out()) cleanup;
	   for p = fpmin to fpmax do
	     output_part.(p - pmin) <- Some writer
	   done;
	   writer
    	)
	t.shuffle_output in

    let output_files =
      List.map
	(fun (filename,fpmin,fpmax) -> [filename, 0L, 0L])
	t.shuffle_output in

    let reader =
      if Array.length input_array > 1 then
	merging_reader input_array extract esys  (* esys: only for to_fd_e *)
      else
	input_array.(0) in

    if t.shuffle_reduce then (
      (** this is a merge+reduce pass *)
      assert(pmin = pmax);
      let writer = List.hd outputs in
      mj # reduce me jc ti reader writer;
    )
    else (
      (** this is a merge+split pass *)
      try
	while true do
	  let line = reader # input_record() in
	  let key = extract line in
	  let p = part_key key in
	  
	  if p < pmin || p > pmax then
	    failwith 
	      (sprintf "Shuffle: Partition outside allowed range \
                          (key=%s p=%d)" key p
	      );
	  
	  ( match output_part.(p - pmin) with
	      | None ->
		  failwith "Shuffle: output partition missing"
	      | Some writer ->
		  writer # output_record line
	  );
	done
      with End_of_file -> ()
    );

    List.iter (fun writer -> writer # close_out()) outputs;
    List.iter (fun reader -> reader # close_in()) inputs;

    if t.shuffle_input_del then (
      List.iter
	(fun (file,_,_) -> delete_files csync file)
	t.shuffle_input
    );

     run_stack (fun f -> f()) cleanup;
    
    `Ok output_files
  with
    | error ->
	let msg =
	  sprintf "Exception %s%s"
	    (Netexn.to_string error)
	    (if Printexc.backtrace_status() then
	       " - backtrace: " ^ (Printexc.get_backtrace())
	     else ""
	    ) in
	run_stack (fun f -> try f() with _ -> ()) cleanup;
	`Error msg


let execute (me:Mapred_def.mapred_env) 
            (jc:Mapred_def.mapred_job_config)
            (mj:Mapred_def.mapred_job) t ti
    : task_result =
  Netlog.logf `Notice "Start task %s"
    (Mapred_tasks.string_of_task_id t);
  let r =
    match t with
      | `Map t' -> 
	  exec_map_task me jc mj t' ti
      | `Sort t' ->
	  exec_sort_task me jc mj t'
      | `Shuffle t' ->
	  exec_shuffle_task me jc mj t' ti
      | `Cleanup ->
	  `Ok []   (* no cleanup here. This is done in mapred_task_server *)
  in
  Netlog.logf `Notice "End task %s"
    (Mapred_tasks.string_of_task_id t);
  r
